/*******************************************************************************
 * Copyright (c) 2015, 2017 IRT AESE (IRT Saint Exupery).
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Bassem Ouni and Pierre Gaufillet (IRT Saint Exupéry) - initial API and implementation
 *     
 *******************************************************************************/
package com.irtsaintexupery.capella2aadl.requests;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.polarsys.capella.common.data.modellingcore.AbstractType;
import org.polarsys.capella.common.helpers.EcoreUtil2;
import org.polarsys.capella.core.data.capellacore.CapellacoreFactory;
import org.polarsys.capella.core.data.capellacore.ModellingArchitecture;
import org.polarsys.capella.core.data.capellacore.StringPropertyValue;
import org.polarsys.capella.core.data.capellamodeller.Project;
import org.polarsys.capella.core.data.capellamodeller.SystemEngineering;
import org.polarsys.capella.core.data.cs.AbstractDeploymentLink;
import org.polarsys.capella.core.data.cs.CsFactory;
import org.polarsys.capella.core.data.cs.DeployableElement;
import org.polarsys.capella.core.data.cs.Part;
import org.polarsys.capella.core.data.cs.PhysicalLink;
import org.polarsys.capella.core.data.cs.PhysicalPort;
import org.polarsys.capella.core.data.fa.ComponentExchange;
import org.polarsys.capella.core.data.fa.ComponentPort;
import org.polarsys.capella.core.data.fa.ComponentPortAllocation;
import org.polarsys.capella.core.data.fa.OrientationPortKind;
import org.polarsys.capella.core.data.pa.PaFactory;
import org.polarsys.capella.core.data.pa.PhysicalArchitecture;
import org.polarsys.capella.core.data.pa.PhysicalComponent;
import org.polarsys.capella.core.data.pa.PhysicalComponentKind;
import org.polarsys.capella.core.data.pa.PhysicalComponentNature;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;

/**
 * This class provides various static methods for transforming Capella physical
 * architecture into a preliminary but yet complete AADL model. The main
 * transformation consists in moving from the 1 to 1 communication model used by
 * {@link ComponentExchange} to a bus oriented topology closer to AADL bus
 * accesses
 * 
 */
public class ComponentExchangeUtils {

	/**
	 * This map provides the list of dummy {@link PhysicalComponent}
	 * representing deployed bahevior nodes for each {@link PhysicalComponent}
	 */
	private static ArrayListMultimap<PhysicalComponent, PhysicalComponent> componentVirtualBehaviorNode;

	private static ArrayListMultimap<PhysicalComponent, PhysicalPort> componentVirtualBehaviorPort;

	private static ArrayListMultimap<PhysicalComponent, PhysicalLink> componentVirtualBehaviorLink;
	private static Multimap<PhysicalLink, PhysicalComponent> virtualBehaviorLinkComponent = null;

	private static ArrayListMultimap<String, PhysicalLink> allocationOfVirtualLinkToPhysicalComponent;
	
	private static Map<ComponentPort, PhysicalPort> allocationOfComponentPortToPhysicalPort;
	
	private static List<PhysicalLink> allocatedVirtualBehaviorLink;

	/**
	 * @param pc
	 * @return
	 */
	public static List<PhysicalComponent> getComponentVirtualBehaviorNode(PhysicalComponent pc) {
		return componentVirtualBehaviorNode.get(pc);
	}

	/**
	 * @param pc
	 * @return
	 */
	public static List<PhysicalPort> getComponentVirtualBehaviorPort(PhysicalComponent pc) {
		return componentVirtualBehaviorPort.get(pc);
	}

	/**
	 * @param pc
	 * @return
	 */
	public static List<PhysicalLink> getComponentVirtualBehaviorLink(PhysicalComponent pc) {
		return componentVirtualBehaviorLink.get(pc);
	}

	/**
	 * @param link
	 * @return
	 */
	public static Collection<PhysicalComponent> getVirtualBehaviorLinkComponent(PhysicalLink link) {
		if(virtualBehaviorLinkComponent == null) {
			virtualBehaviorLinkComponent = Multimaps.invertFrom(componentVirtualBehaviorLink, ArrayListMultimap.<PhysicalLink, PhysicalComponent>create());
//			System.out.println("Reversing component/link map...");
//			for(PhysicalLink l: virtualBehaviorLinkComponent.keySet()) {
//				System.out.println(l.getName() + ":");
//				for(PhysicalComponent c: virtualBehaviorLinkComponent.get(l)) {
//					System.out.println("\t" + c.getName());
//				}
//			}
		}
		return virtualBehaviorLinkComponent.get(link);
	}

	/**
	 * @param pc
	 * @return
	 */
	public static List<PhysicalLink> getAllocatedVirtualBehaviorLink() {
		return allocatedVirtualBehaviorLink;
	}

	public static void transformCapellaComponentExchangeIntoAADLConnections(Project project) throws Capella2AADLException {
		// Get the physical architecture from Project in Capella model
		List<ModellingArchitecture> architectures = ((SystemEngineering) (project.getOwnedModelRoots().get(0))).getOwnedArchitectures();
		PhysicalArchitecture physicalArchitecture = null;

		for (ModellingArchitecture architecture : architectures) {
			if (architecture instanceof PhysicalArchitecture) {
				physicalArchitecture = (PhysicalArchitecture) architecture;
				break;
			}
		}

		// If no physical architecture has been found, the transformation
		// doesn't apply
		if (physicalArchitecture == null) {
			throw new Capella2AADLException("Physical Architecture not found.");
		}

		PhysicalComponent physicalSystem = physicalArchitecture.getOwnedPhysicalComponent();

		// Index the PhysicalComponent parts (nature NODE, kind
		// SOFTWARE_EXECUTION_UNIT)

		System.out.println("CPU/Device parts:");
		List<Part> hardwareParts = new ArrayList<>();
		
		allocationOfComponentPortToPhysicalPort = new HashMap<>();
		
		for (Iterator<EObject> it = physicalSystem.eAllContents(); it.hasNext();) {
			EObject eo = it.next();
			if (eo instanceof Part) {
				Part part = (Part) eo;
				AbstractType abstractType = part.getAbstractType();
				if (abstractType instanceof PhysicalComponent) {
					PhysicalComponent typeComponent = (PhysicalComponent) abstractType;
					if ((typeComponent.getKind().equals(PhysicalComponentKind.SOFTWARE_EXECUTION_UNIT) || typeComponent.getKind().equals(
							PhysicalComponentKind.HARDWARE))
							&& typeComponent.getNature().equals(PhysicalComponentNature.NODE)) {
						hardwareParts.add(part);
						System.out.println("\t" + PhysicalLinkUtils.getFullQualifiedName(part) + " (" + part.getAbstractType().getName() + ")");
					}
				}
			}
			if(eo instanceof ComponentPortAllocation) {
				ComponentPortAllocation cpa = (ComponentPortAllocation) eo;
				allocationOfComponentPortToPhysicalPort.put((ComponentPort) cpa.getTargetElement(), (PhysicalPort) cpa.getSourceElement());
			}
		}

		System.out.println();
		System.out.println("Deployed components:");
		// Build a Behavioral node to CPU/device part map
		Map<PhysicalComponent, Part> behaviourNode2Part = new HashMap<>();
		for (Part part : hardwareParts) {
			for (AbstractDeploymentLink deployableLink : part.getOwnedDeploymentLinks()) {
				DeployableElement deployableElement = deployableLink.getDeployedElement();
				if (deployableElement instanceof Part) {
					Part deployedPart = (Part) deployableElement;
					AbstractType deployedPartType = deployedPart.getAbstractType();
					if (deployedPartType instanceof PhysicalComponent) {
						PhysicalComponent deployedPhysicalComponent = (PhysicalComponent) deployedPartType;
						behaviourNode2Part.put(deployedPhysicalComponent, part);
						System.out.println("\t" + deployedPhysicalComponent.getName() + " is deployed on " + PhysicalLinkUtils.getFullQualifiedName(part));
					}
				}
			}
		}

		System.out.println();
		componentVirtualBehaviorNode = ArrayListMultimap.create();
		componentVirtualBehaviorPort = ArrayListMultimap.create();
		componentVirtualBehaviorLink = ArrayListMultimap.create();
		allocationOfVirtualLinkToPhysicalComponent = ArrayListMultimap.create();
		allocatedVirtualBehaviorLink = new ArrayList<>();
		virtualBehaviorLinkComponent = null;
		
		// Fow now let's consider ComponentExchanges are all in the root
		// physical system
		for (ComponentExchange ce : physicalSystem.getOwnedComponentExchanges()) {
			// First determine which PhysicalComponent is the virtual parent of
			// the ports' containers
			ComponentPort sourcePort = (ComponentPort) ce.getSource();
			ComponentPort targetPort = (ComponentPort) ce.getTarget();

			// Re-orient the link if needed. The sourcePort must really be the sender
			if(sourcePort.getOrientation().equals(OrientationPortKind.IN)) {
				ComponentPort tmpPort = targetPort;
				targetPort = sourcePort;
				sourcePort = tmpPort;
			}
			
			boolean physicalComponentCleanUpNeeded = false;

			System.out.println("Processing ComponentExchange " + PhysicalLinkUtils.getFullQualifiedName(sourcePort) + " <-> "
					+ PhysicalLinkUtils.getFullQualifiedName(targetPort));

			PhysicalComponent sourcePortContainer = (PhysicalComponent) sourcePort.eContainer();
			PhysicalComponent targetPortContainer = (PhysicalComponent) targetPort.eContainer();

			System.out.println("\tsourcePortContainer: " + PhysicalLinkUtils.getFullQualifiedName(sourcePortContainer));
			System.out.println("\ttargetPortContainer: " + PhysicalLinkUtils.getFullQualifiedName(targetPortContainer));
			// Search now the PhysicalComponent part (nature NODE, kind
			// SOFTWARE_EXECUTION_UNIT) referencing these containers

			// The parent of this part/type is the virtual parent of the
			// ports' containers
			Part virtualSourcePortParentPart = behaviourNode2Part.get(sourcePortContainer);
			Part virtualTargetPortParentPart = behaviourNode2Part.get(targetPortContainer);
			PhysicalComponent virtualSourcePortGrandParent = null;
			PhysicalComponent virtualTargetPortGrandParent = null;
			if (virtualSourcePortParentPart != null) {
				switch (((PhysicalComponent) behaviourNode2Part.get(sourcePortContainer).getAbstractType()).getKind()) {
				case SOFTWARE_EXECUTION_UNIT: {
					virtualSourcePortGrandParent = (PhysicalComponent) behaviourNode2Part.get(sourcePortContainer).eContainer();
					break;
				}
				case HARDWARE: {
					virtualSourcePortGrandParent = (PhysicalComponent) behaviourNode2Part.get(sourcePortContainer).getAbstractType();
					break;
				}
				default: {
					throw new Capella2AADLException("Unknown PhysicalComponent kind");
				}
				}
			}
			if (virtualTargetPortParentPart != null) {
				switch (((PhysicalComponent) behaviourNode2Part.get(targetPortContainer).getAbstractType()).getKind()) {
				case SOFTWARE_EXECUTION_UNIT: {
					virtualTargetPortGrandParent = (PhysicalComponent) behaviourNode2Part.get(targetPortContainer).eContainer();
					break;
				}
				case HARDWARE: {
					virtualTargetPortGrandParent = (PhysicalComponent) behaviourNode2Part.get(targetPortContainer).getAbstractType();
					break;
				}
				default: {
					throw new Capella2AADLException("Unknown PhysicalComponent kind");
				}
				}
			}

			if (virtualSourcePortGrandParent != null && virtualTargetPortGrandParent != null) {

				System.out.println("\tvirtualSourcePortGrandParent: " + PhysicalLinkUtils.getFullQualifiedName(virtualSourcePortGrandParent));
				System.out.println("\tvirtualTargetPortGrandParent: " + PhysicalLinkUtils.getFullQualifiedName(virtualTargetPortGrandParent));

				PhysicalComponent virtualSourcePortParent = null;
				PhysicalComponent virtualTargetPortParent = null;
				switch (((PhysicalComponent) behaviourNode2Part.get(sourcePortContainer).getAbstractType()).getKind()) {
				case SOFTWARE_EXECUTION_UNIT: {
					virtualSourcePortParent = PaFactory.eINSTANCE.createPhysicalComponent(sourcePortContainer.getName());
					componentVirtualBehaviorNode.put(virtualSourcePortGrandParent, virtualSourcePortParent);
					virtualSourcePortGrandParent.getOwnedPhysicalComponents().add(virtualSourcePortParent);
					physicalComponentCleanUpNeeded = true;
					break;
				}
				case HARDWARE: {
					// Connect directly to the device
					virtualSourcePortParent = virtualSourcePortGrandParent;
					break;
				}
				default: {
					throw new Capella2AADLException("Unknown PhysicalComponent kind");
				}
				}

				switch (((PhysicalComponent) behaviourNode2Part.get(targetPortContainer).getAbstractType()).getKind()) {
				case SOFTWARE_EXECUTION_UNIT: {
					virtualTargetPortParent = PaFactory.eINSTANCE.createPhysicalComponent(targetPortContainer.getName());
					componentVirtualBehaviorNode.put(virtualTargetPortGrandParent, virtualTargetPortParent);
					virtualTargetPortGrandParent.getOwnedPhysicalComponents().add(virtualTargetPortParent);
					physicalComponentCleanUpNeeded = true;
					break;
				}
				case HARDWARE: {
					// Connect directly to the device
					virtualTargetPortParent = virtualTargetPortGrandParent;
					break;
				}

				default: {
					throw new Capella2AADLException("Unknown PhysicalComponent kind");
				}
				}

				// Add tip ports
				// And the property "orientation" provides the orientation of
				// the port
				StringPropertyValue outOrientationProperty = CapellacoreFactory.eINSTANCE.createStringPropertyValue("orientation");
				outOrientationProperty.setValue("out");
				StringPropertyValue inOrientationProperty = CapellacoreFactory.eINSTANCE.createStringPropertyValue("orientation");
				inOrientationProperty.setValue("in");

				PhysicalPort virtualSourcePort = CsFactory.eINSTANCE.createPhysicalPort(sourcePort.getName());
				PhysicalPort virtualTargetPort = CsFactory.eINSTANCE.createPhysicalPort(targetPort.getName());
				virtualSourcePort.getOwnedPropertyValues().add(outOrientationProperty);
				virtualTargetPort.getOwnedPropertyValues().add(inOrientationProperty);

				componentVirtualBehaviorPort.put(virtualSourcePortParent, virtualSourcePort);
				componentVirtualBehaviorPort.put(virtualTargetPortParent, virtualTargetPort);

				// BEWARE: here the orginal PhysicalComponent representing a
				// device is directly modified.
				// This strategy would impact the bus topology transformation.
				// Therefore, the ComponentExchange processing
				// shall be done AFTER the bus topology.
				virtualSourcePortParent.getOwnedFeatures().add(virtualSourcePort);
				virtualTargetPortParent.getOwnedFeatures().add(virtualTargetPort);

				// Create virtualLink
				PhysicalLink link = CsFactory.eINSTANCE.createPhysicalLink(ce.getName());
				link.getLinkEnds().add(virtualSourcePort);
				link.getLinkEnds().add(virtualTargetPort);

				// Extracting allocation to bus
				Set<Bus> allocatingBuses = new HashSet<Bus>();
				Bus lastAllocatingBus = null;
				Bus tmpBus = PhysicalLinkUtils.getBusForPort(allocationOfComponentPortToPhysicalPort.get(sourcePort));
				if (tmpBus != null) {
					lastAllocatingBus = tmpBus;
					allocatingBuses.add(lastAllocatingBus);
				}
				tmpBus = PhysicalLinkUtils.getBusForPort(allocationOfComponentPortToPhysicalPort.get(targetPort));
				if (tmpBus != null) {
					lastAllocatingBus = tmpBus;
					allocatingBuses.add(lastAllocatingBus);
				}		
				
				if(allocatingBuses.size() > 1)
				{
					// Error : a link cannot be allocated to several buses
					throw new Capella2AADLException("Ports " + PhysicalLinkUtils.getFullQualifiedName(sourcePort) + " and " + PhysicalLinkUtils.getFullQualifiedName(targetPort) + " shall be allocated to the same bus");
				}
				
				if(allocatingBuses.isEmpty())
				{
					// Error : a link shall be allocated to one bus
					throw new Capella2AADLException("Ports " + PhysicalLinkUtils.getFullQualifiedName(sourcePort) + " or " + PhysicalLinkUtils.getFullQualifiedName(targetPort) + " shall be allocated to a bus");
				}

				synthesizePortsAndLinks(link, lastAllocatingBus);

				if (physicalComponentCleanUpNeeded) {
					virtualTargetPortGrandParent.getOwnedPhysicalComponents().remove(virtualTargetPortParent);
					virtualSourcePortGrandParent.getOwnedPhysicalComponents().remove(virtualSourcePortParent);
				}
			}
		}
	}

	private static void synthesizePortsAndLinks(PhysicalLink link, Bus bus) throws Capella2AADLException {

		// This algorithm only works if link is between a tip port and the bus
		// port.
		// In this situation, the bus is always contained in the closest common
		// ancestor of all tip ports and links are always exclusively ascending
		// in the hierarchy.

		// Note: in Capella models, the grand parent of a PhysicalPort cannot be
		// null
		PhysicalPort sourcePort, targetPort;

		// Origin and destination ports (links are not oriented)
		sourcePort = (PhysicalPort) link.getLinkEnds().get(0);
		targetPort = (PhysicalPort) link.getLinkEnds().get(1);

		System.out.println(PhysicalLinkUtils.getFullQualifiedName(sourcePort) + " <-> " + PhysicalLinkUtils.getFullQualifiedName(targetPort));

		PhysicalPort lastPort1 = createBottomUpLinks(sourcePort, targetPort, link.getName(), OrientationPortKind.OUT);
		PhysicalPort lastPort2 = createBottomUpLinks(targetPort, sourcePort, link.getName(), OrientationPortKind.IN);

		// Now take care of the final case where ancestor = busPortGrandParent.
		// Here ports have already been created and it may only be needed to
		// connect ports. The algorithm is exactly the same as above.
		
		createSyntheticLink(link.getName(), lastPort1, lastPort2, (PhysicalComponent) EcoreUtil2.getCommonAncestor(sourcePort, targetPort), bus);
	}

	public static PhysicalPort createBottomUpLinks(PhysicalPort sourcePort, PhysicalPort targetport, String name, OrientationPortKind orientation) throws Capella2AADLException {
		PhysicalComponent ancestor = (PhysicalComponent) sourcePort.eContainer().eContainer();
		PhysicalComponent targetPortGrandParent = (PhysicalComponent) targetport.eContainer().eContainer();
		PhysicalPort origin = sourcePort;
		PhysicalPort syntheticPhysicalPort;

		while (!EcoreUtil.isAncestor(ancestor, targetPortGrandParent)) {
			syntheticPhysicalPort = CsFactory.eINSTANCE.createPhysicalPort(name);
			componentVirtualBehaviorPort.put(ancestor, syntheticPhysicalPort);
			ancestor.getOwnedFeatures().add(syntheticPhysicalPort);
			// And the property "orientation" provides the orientation of the
			// port
			StringPropertyValue orientationProperty = CapellacoreFactory.eINSTANCE.createStringPropertyValue("orientation");
			switch(orientation) {
			case IN:
				orientationProperty.setValue("in");
				createSyntheticLink(name, syntheticPhysicalPort, origin, ancestor);
				break;
			case OUT: 
				orientationProperty.setValue("out");
				createSyntheticLink(name, origin, syntheticPhysicalPort, ancestor);
				break;
			default: 
				throw new Capella2AADLException("Unknown orientation");
			}
			syntheticPhysicalPort.getOwnedPropertyValues().add(orientationProperty);

			System.out.println("Create port " + PhysicalLinkUtils.getFullQualifiedName(syntheticPhysicalPort));

			// Move the origin to the syntheticPort, reset syntheticPort and
			// move ancestor one step upward in the hierarchy
			origin = syntheticPhysicalPort;
			ancestor = (PhysicalComponent) ancestor.eContainer();
		}
		return origin;
	}

	/**
	 * This method creates a {@link PhysicalLink} between {@link PhysicalPort}
	 * <code>origin</code> and <code>destination</code> and stores it in
	 * {@link PhysicalComponent} <code>ancestor</code> only if no such link
	 * already exists.
	 * 
	 * @param prefix
	 *            Name of the new link
	 * @param origin
	 *            Port of origin
	 * @param destination
	 *            Port of destination
	 * @param ancestor
	 *            Container of the new link
	 */
	private static void createSyntheticLink(String prefix, PhysicalPort origin, PhysicalPort destination, PhysicalComponent ancestor) {
		createSyntheticLink(prefix, origin, destination, ancestor, null);
	}

	private static void createSyntheticLink(String prefix, PhysicalPort origin, PhysicalPort destination, PhysicalComponent ancestor, Bus allocatedTo) {
		boolean needToCreateNewLink = true;
		for (PhysicalLink l : componentVirtualBehaviorLink.get(ancestor)) {
			if ((l.getLinkEnds().get(0).equals(destination) && l.getLinkEnds().get(1).equals(origin))
					|| (l.getLinkEnds().get(0).equals(origin) && l.getLinkEnds().get(1).equals(destination))) {
				needToCreateNewLink = false;
				break;
			}
		}

		if (needToCreateNewLink) {
			PhysicalLink syntheticPhysicalLink = CsFactory.eINSTANCE.createPhysicalLink(prefix + "_" + (componentVirtualBehaviorLink.get(ancestor).size() + 1));
			syntheticPhysicalLink.getLinkEnds().add(origin);
			syntheticPhysicalLink.getLinkEnds().add(destination);
			componentVirtualBehaviorLink.put(ancestor, syntheticPhysicalLink);
			
			if(allocatedTo != null) {
			// The property "allocatedToBus" provides the name of the bus instance the link is allocated to
			StringPropertyValue busName = CapellacoreFactory.eINSTANCE.createStringPropertyValue("allocatedToBus");
			busName.setValue(allocatedTo.getName());
			syntheticPhysicalLink.getOwnedPropertyValues().add(busName);
			allocationOfVirtualLinkToPhysicalComponent.put(allocatedTo.getName(), syntheticPhysicalLink);
			allocatedVirtualBehaviorLink.add(syntheticPhysicalLink);
			}
			
			System.out.println("Create synthetic link: " + PhysicalLinkUtils.getFullQualifiedName(origin) + " <-> "
					+ PhysicalLinkUtils.getFullQualifiedName(destination));
		}
	}

}
