/*******************************************************************************
 * Copyright (c) 2015, 2017 IRT AESE (IRT Saint Exupery).
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Bassem Ouni and Pierre Gaufillet (IRT Saint Exupéry) - initial API and implementation
 *     
 *******************************************************************************/
package com.irtsaintexupery.capella2aadl.requests;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.polarsys.capella.common.helpers.EcoreUtil2;
import org.polarsys.capella.core.data.capellacore.AbstractPropertyValue;
import org.polarsys.capella.core.data.capellacore.BooleanPropertyValue;
import org.polarsys.capella.core.data.capellacore.CapellaElement;
import org.polarsys.capella.core.data.capellacore.CapellacoreFactory;
import org.polarsys.capella.core.data.capellacore.Feature;
import org.polarsys.capella.core.data.capellacore.ModellingArchitecture;
import org.polarsys.capella.core.data.capellacore.NamedElement;
import org.polarsys.capella.core.data.capellacore.StringPropertyValue;
import org.polarsys.capella.core.data.capellamodeller.Project;
import org.polarsys.capella.core.data.capellamodeller.SystemEngineering;
import org.polarsys.capella.core.data.cs.AbstractPhysicalLinkEnd;
import org.polarsys.capella.core.data.cs.CsFactory;
import org.polarsys.capella.core.data.cs.PhysicalLink;
import org.polarsys.capella.core.data.cs.PhysicalPort;
import org.polarsys.capella.core.data.pa.PaFactory;
import org.polarsys.capella.core.data.pa.PhysicalArchitecture;
import org.polarsys.capella.core.data.pa.PhysicalComponent;
import org.polarsys.capella.core.data.pa.PhysicalComponentNature;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;

/**
 * This class provides various static methods for transforming Capella physical
 * architecture into a preliminary but yet complete AADL model. The main
 * transformation consists in moving from the 1 to 1 communication model used by
 * {@link PhysicalLink} to a bus oriented topology closer to AADL bus accesses
 * 
 */
public class PhysicalLinkUtils {

	/**
	 * Reversible map between deduced buses and origin/destination
	 * {@link PhysicalPort}
	 */
	private static BiMap<Bus, Set<PhysicalPort>> bus2TipPorts = null;

	/**
	 * Cross-reference from {@link PhysicalPort} to {@link PhysicalLink}
	 * connected to these ports
	 */
	private static ArrayListMultimap<PhysicalPort, PhysicalLink> port2link = null;

	/**
	 * All {@link PhysicalLink} in the model
	 */
	private static List<PhysicalLink> physicalLinks = null;

	/**
	 * List of {@link PhysicalPort} to synthesize for each
	 * {@link PhysicalComponent}
	 */
	private static ArrayListMultimap<PhysicalComponent, PhysicalPort> componentSyntheticPorts = null;

	/**
	 * Cross-reference from {@link PhysicalPort} to the bus they are connected
	 * to
	 */
	private static Map<PhysicalPort, Bus> componentSyntheticPorts2Bus = null;

	/**
	 * List of {@link PhysicalLink} to synthesize for each
	 * {@link PhysicalComponent}
	 */
	private static ArrayListMultimap<PhysicalComponent, PhysicalLink> componentSyntheticLinks = null;

	/**
	 * Numbering counter for default bus types
	 */
	private static int defaultBusTypeCounter;

	/**
	 * @param pc
	 *            target {@link PhysicalComponent}
	 * @return the list of synthetic ports for component <code>pc</code>
	 */
	public static List<PhysicalPort> getSyntheticPorts(PhysicalComponent pc) {
		return componentSyntheticPorts.get(pc);
	}

	/**
	 * @param pc
	 *            target {@link PhysicalComponent}
	 * @return the list of synthetic links for component <code>pc</code>
	 */
	public static List<PhysicalLink> getSyntheticLinks(PhysicalComponent pc) {
		return componentSyntheticLinks.get(pc);
	}

	/**
	 * @param port
	 *            target {@link PhysicalPort}
	 * @return the type name of the bus connected to <code>port</code>
	 */
	public static String getBusTypeForPort(PhysicalPort port) {
		return componentSyntheticPorts2Bus.get(port).getType();
	}

	/**
	 * @param port
	 *            target {@link PhysicalPort}
	 * @return the bus connected to <code>port</code>
	 */
	public static Bus getBusForPort(PhysicalPort port) {
		return componentSyntheticPorts2Bus.get(port);
	}

	/**
	 * This methode analyses the data flows and hierarchy of
	 * {@link PhysicalComponent} of the physical architecture and synthesizes a
	 * corresponding bus topology.
	 * 
	 * @param project
	 *            Capella {@link Project} that is analyzed
	 * @throws Capella2AADLException
	 */
	public static void transformCapellaPhysicalLinksIntoAADLBusTopology(Project project) throws Capella2AADLException {

		// (Re-)init Bus class attributes and bus type counter
		Bus.init();
		defaultBusTypeCounter = 1;
		
		// Get the physical architecture from Project in Capella model
		List<ModellingArchitecture> architectures = ((SystemEngineering) (project.getOwnedModelRoots().get(0))).getOwnedArchitectures();
		PhysicalArchitecture physicalArchitecture = null;

		for (ModellingArchitecture architecture : architectures) {
			if (architecture instanceof PhysicalArchitecture) {
				physicalArchitecture = (PhysicalArchitecture) architecture;
				break;
			}
		}

		// If no physical architecture has been found, the transformation
		// doesn't apply
		if (physicalArchitecture == null) {
			throw new Capella2AADLException("Physical Architecture not found.");
		}

		PhysicalComponent physicalSystem = physicalArchitecture.getOwnedPhysicalComponent();

		// Visit the model and connect internally PhysicalPorts of PhysicalComponents
		// when they share the same virtualConnection string property value
		// Thanks to the preprocessing, it is possible to consider low level
		// communication device like transceivers, routers, switch, etc. as physically connecting 
		// some of their ports.
		
		for (Iterator<EObject> it = physicalSystem.eAllContents(); it.hasNext();) {
			EObject eo = it.next();
			if (eo instanceof PhysicalComponent) {
				PhysicalComponent pc = (PhysicalComponent) eo;
				ArrayListMultimap<String, PhysicalPort> map = ArrayListMultimap.create();
				for(Feature feature: pc.getOwnedFeatures()) {
					if (feature instanceof PhysicalPort) {
						PhysicalPort pp = (PhysicalPort) feature;
						for(AbstractPropertyValue pv: pp.getAppliedPropertyValues()) {
							if(pv.getName().equals("virtualConnection")) {
								map.put(((StringPropertyValue) pv).getValue(), pp);
							}
						}
					}
				}
				// Now that the list of ports is known, create a chain with all ports of each virtualConnection
				for(String virtualConnection: map.keySet()) {
					int counter = 1;
					Iterator<PhysicalPort> pp_it = map.get(virtualConnection).iterator();
					PhysicalPort origin = pp_it.next();
					PhysicalPort dest = null;
					PhysicalLink link = null;
					while(pp_it.hasNext()) {
						dest = pp_it.next();
						link = CsFactory.eINSTANCE.createPhysicalLink(virtualConnection + "_" + counter);
						link.getLinkEnds().add(origin);
						link.getLinkEnds().add(dest);
						PhysicalComponent parentComponent = (PhysicalComponent) origin.eContainer();
						parentComponent.getOwnedPhysicalLinks().add(link);
						System.out.println("Creating virtual connection "
								+ virtualConnection + "_" + counter + " in "
								+ getFullQualifiedName(parentComponent) + ": "
								+ origin.getName() + " <-> " + dest.getName());
						if(pp_it.hasNext()) {
							origin = dest;
						}
					}
				}
			}
		}
		
		// Extract the list of physical ports, physical links and a cross
		// reference between
		// physical ports and physical links

		physicalLinks = new ArrayList<PhysicalLink>();
		port2link = ArrayListMultimap.create();

		for (Iterator<EObject> it = physicalSystem.eAllContents(); it.hasNext();) {
			EObject eo = it.next();
			if (eo instanceof PhysicalLink) {
				PhysicalLink link = (PhysicalLink) eo;
				physicalLinks.add(link);
				port2link.put((PhysicalPort) link.getLinkEnds().get(0), link);
				port2link.put((PhysicalPort) link.getLinkEnds().get(1), link);
			}
		}

		// Now extract the origin and destination ports (aka tipPorts) for each
		// set of links (PhysicalLink) connected through ports (PhysicalPort)

		bus2TipPorts = HashBiMap.create();

		// Each PhysicalLink shall be visited once. Initialize the list of links
		// to visit
		List<PhysicalLink> linksToVisit = new ArrayList<PhysicalLink>(physicalLinks);

		// allConnectedLinks is a list of sets of PhysicalLink connected by
		// PhysicalPorts. Sets of links are never connected. Each set will lead
		// to an AADL bus instance
		ArrayDeque<Set<PhysicalLink>> allConnectedLinks = new ArrayDeque<Set<PhysicalLink>>();

		Set<PhysicalPort> ports;
		while (!linksToVisit.isEmpty()) {
			ports = new HashSet<>();
			allConnectedLinks.add(extractConnectedPhysicalLinks(linksToVisit.get(0), ports));

			// Now that connected links and ports have been discovered, identify
			// the type of the bus. By convention, this type is hold by a
			// PhysicalLinkCategory associated with at least one connected link
			String categoryName = null;
			for (PhysicalLink l : allConnectedLinks.peekLast()) {
				if (!l.getCategories().isEmpty()) {
					categoryName = l.getCategories().get(0).getName();
					break;
				}
			}

			// If no category has been found, use a default name
			if (categoryName == null) {
				categoryName = "BusType_" + defaultBusTypeCounter;
				defaultBusTypeCounter++;
			}

			// Gather all these information together
			// the bus instance is named by its category + an increment
			Bus bus = Bus.createIncrementedNameFromCategory(categoryName);

			// Associate the bus with its tip ports
			bus2TipPorts.put(bus, ports);

			// all the connected links has been visited in the process. They can
			// be removed at once from the list of links that still need to be
			// visited.
			linksToVisit.removeAll(allConnectedLinks.peekLast());
		}

		printTipPorts2Bus();

		// Now that the original topology of PhysicalLinks is known, synthesize
		// a new one compatible with AADL. By convention, a new
		// PhysicalComponent with specific key/values will be created in the
		// lowest ancestor of all tip ports.

		componentSyntheticPorts = ArrayListMultimap.create();
		componentSyntheticLinks = ArrayListMultimap.create();

		// Initialize the list of synthetic ports with the tip ports
		// Only one port / component / bus should be kept
		componentSyntheticPorts2Bus = new HashMap<>();
		for (Set<PhysicalPort> sp : bus2TipPorts.inverse().keySet()) {
			for (PhysicalPort p : sp) {
				if (!componentSyntheticPorts.containsValue(p)) {
					componentSyntheticPorts2Bus.put(p, bus2TipPorts.inverse().get(sp));
					componentSyntheticPorts.put((PhysicalComponent) p.eContainer(), p);
				}
			}
		}

		for (Bus bus : bus2TipPorts.keySet()) {
			// First discover the closest common ancestor of all tipPorts
			PhysicalComponent commonAncestor = (PhysicalComponent) EcoreUtil2.getCommonAncestor(bus2TipPorts.get(bus));

			// Create the PhysicalComponent that will play the role of bus
			// instance
			PhysicalComponent busPhysicalComponent = PaFactory.eINSTANCE.createPhysicalComponent(bus.getType());
			busPhysicalComponent.setNature(PhysicalComponentNature.NODE);

			// The property "isBus" acts as a marker to process differently this
			// PhysicalComponent in Acceleo
			BooleanPropertyValue isBus = CapellacoreFactory.eINSTANCE.createBooleanPropertyValue("isBus");
			isBus.setValue(true);
			busPhysicalComponent.getOwnedPropertyValues().add(isBus);

			// And the property "busName" provides the name of the bus instance
			StringPropertyValue busName = CapellacoreFactory.eINSTANCE.createStringPropertyValue("busName");
			busName.setValue(bus.getName());
			busPhysicalComponent.getOwnedPropertyValues().add(busName);

			// Add a port to the bus instance to connect synthetic links to it
			PhysicalPort port = CsFactory.eINSTANCE.createPhysicalPort("bus_port");
			busPhysicalComponent.getOwnedFeatures().add(port);

			// And finally add the bus instance to the common ancestor
			commonAncestor.getOwnedPhysicalComponents().add(busPhysicalComponent);

			// Synthesize now the new PhysicalLinks. It is done in 2 steps:
			// 1- Create a link from tip port to bus_port
			// 2- Split this link into segments that don't cross
			// PhysicalComponent boundaries
			// The second step involves synthesize intermediate PhysicalPorts on
			// intermediate PhysicalComponents
			for (PhysicalPort p : bus2TipPorts.get(bus)) {
				PhysicalLink l = CsFactory.eINSTANCE.createPhysicalLink(bus.getName());
				l.getLinkEnds().add(p);
				l.getLinkEnds().add(port);
				synthesizePortsAndLinks(l, bus);
			}
		}

		System.out.println();
		printSyntheticPorts();
		printSyntheticLinks();
		System.out.println();
	}

	/**
	 * Print synthetic links by {@link PhysicalComponent} on stdout for
	 * debugging purpose
	 */
	private static void printSyntheticLinks() {
		System.out.println("\nSynthetic Links:");
		for (PhysicalComponent c : componentSyntheticLinks.keySet()) {
			System.out.println(getFullQualifiedName(c) + ":");
			for (PhysicalLink l : componentSyntheticLinks.get(c)) {
				System.out.println("\t" + getFullQualifiedName(l.getLinkEnds().get(0)) + " <-> " + getFullQualifiedName(l.getLinkEnds().get(1)));
			}
		}
	}

	/**
	 * Print synthetic ports by {@link PhysicalComponent} on stdout for
	 * debugging purpose
	 */
	private static void printSyntheticPorts() {
		System.out.println("Synthetic Ports:");
		for (PhysicalComponent c : componentSyntheticPorts.keySet()) {
			System.out.println(getFullQualifiedName(c) + ":");
			for (PhysicalPort p : componentSyntheticPorts.get(c)) {
				System.out.println("\t" + getFullQualifiedName(p));
			}
		}
	}

	/**
	 * This method analyzes the {@link PhysicalLink} link and if required splits
	 * it into several links connected by new intermediate ports that don't
	 * violate the encapsulation principle of AADL.
	 * 
	 * @param link
	 *            the link that will be analyzed
	 * @param bus
	 *            the bus <code>link</code> connects to
	 */
	private static void synthesizePortsAndLinks(PhysicalLink link, Bus bus) {

		// This algorithm only works if link is between a tip port and the bus
		// port.
		// In this situation, the bus is always contained in the closest common
		// ancestor of all tip ports and links are always exclusively ascending
		// in the hierarchy.

		// Note: in Capella models, the grand parent of a PhysicalPort cannot be
		// null
		PhysicalPort tipPort, bus_port;

		// Origin and destination ports (links are not oriented)
		tipPort = (PhysicalPort) link.getLinkEnds().get(0);
		bus_port = (PhysicalPort) link.getLinkEnds().get(1);

		System.out.println(getFullQualifiedName(tipPort) + " <-> " + getFullQualifiedName(bus_port));

		PhysicalComponent ancestor = (PhysicalComponent) tipPort.eContainer().eContainer();
		PhysicalComponent busPortGrandParent = (PhysicalComponent) bus_port.eContainer().eContainer();
		PhysicalPort origin = tipPort;
		PhysicalPort syntheticPhysicalPort = null;

		while (!ancestor.equals(busPortGrandParent)) {

			// Ancestor needs one and only one Physical port to connect to the
			// bus. If a such port already exists, use it rather than create a
			// new one.
			for (PhysicalPort p : componentSyntheticPorts.get(ancestor)) {
				if (componentSyntheticPorts2Bus.containsKey(p)) {
					if (componentSyntheticPorts2Bus.get(p).equals(bus)) {
						syntheticPhysicalPort = p;
						System.out.println("Component " + getFullQualifiedName(ancestor) + " already contains the synthetic port " + getFullQualifiedName(p)
								+ " for bus " + bus.getName());
						break;
					}
				}
			}

			// No port connected to the bus has been found. Create a new one.
			if (syntheticPhysicalPort == null) {
				syntheticPhysicalPort = CsFactory.eINSTANCE.createPhysicalPort(link.getName());
				componentSyntheticPorts.put(ancestor, syntheticPhysicalPort);
				componentSyntheticPorts2Bus.put(syntheticPhysicalPort, bus);
				ancestor.getOwnedFeatures().add(syntheticPhysicalPort);
				System.out.println("Component " + getFullQualifiedName(ancestor) + " doesn't contain a synthetic port for bus " + bus.getName()
						+ ": create port " + getFullQualifiedName(syntheticPhysicalPort));
			}

			createSyntheticLink(link.getName(), origin, syntheticPhysicalPort, ancestor);

			// Move the origin to the syntheticPort, reset syntheticPort and
			// move ancestor one step upward in the hierarchy
			origin = syntheticPhysicalPort;
			syntheticPhysicalPort = null;
			ancestor = (PhysicalComponent) ancestor.eContainer();
		}

		// Now take care of the final case where ancestor = busPortGrandParent.
		// Here ports have already been created and it may only be needed to
		// connect ports. The algorithm is exactly the same as above.
		createSyntheticLink(link.getName(), origin, bus_port, ancestor);
	}

	/**
	 * This method creates a {@link PhysicalLink} between {@link PhysicalPort}
	 * <code>origin</code> and <code>destination</code> and stores it in
	 * {@link PhysicalComponent} <code>ancestor</code> only if no such link
	 * already exists.
	 * 
	 * @param prefix
	 *            Name of the new link
	 * @param origin
	 *            Port of origin
	 * @param destination
	 *            Port of destination
	 * @param ancestor
	 *            Container of the new link
	 */
	private static void createSyntheticLink(String prefix, PhysicalPort origin, PhysicalPort destination, PhysicalComponent ancestor) {
		boolean needToCreateNewLink = true;
		for (PhysicalLink l : componentSyntheticLinks.get(ancestor)) {
			if ((l.getLinkEnds().get(0).equals(destination) && l.getLinkEnds().get(1).equals(origin))
					|| (l.getLinkEnds().get(0).equals(origin) && l.getLinkEnds().get(1).equals(destination))) {
				needToCreateNewLink = false;
				break;
			}
		}

		if (needToCreateNewLink) {
			PhysicalLink syntheticPhysicalLink = CsFactory.eINSTANCE.createPhysicalLink(prefix + "_" + (componentSyntheticLinks.get(ancestor).size() + 1));
			syntheticPhysicalLink.getLinkEnds().add(origin);
			syntheticPhysicalLink.getLinkEnds().add(destination);
			componentSyntheticLinks.put(ancestor, syntheticPhysicalLink);
			System.out.println("Create synthetic link: " + getFullQualifiedName(origin) + " <-> " + getFullQualifiedName(destination));
		}
	}

	/**
	 * Print origin/destination ports by bus on stdout for debugging purpose
	 */
	private static void printTipPorts2Bus() {
		for (Bus b : bus2TipPorts.keySet()) {
			System.out.println(b.getName() + ": " + b.getType());
			for (PhysicalPort p : bus2TipPorts.get(b))
				System.out.println("\t" + getFullQualifiedName(p));
			System.out.println("");
		}
	}

	/**
	 * This method returns the list of {@link PhysicalLink} connected by
	 * {@link PhysicalPort} from <code>link</code> along with the list of
	 * origin/destination ports.
	 * 
	 * @param link
	 *            the link that is analyzed
	 * @param tipPorts
	 *            the list of origin/destination ports
	 * @return
	 */
	private static Set<PhysicalLink> extractConnectedPhysicalLinks(PhysicalLink link, Set<PhysicalPort> tipPorts) {

		Set<PhysicalLink> connectedLinks = new HashSet<PhysicalLink>();
		extractConnectedPhysicalLinks(link, connectedLinks, tipPorts);
		return connectedLinks;
	}

	/**
	 * This method returns the list of {@link PhysicalLink} connected by
	 * {@link PhysicalPort} from <code>link</code> and updates the list of
	 * origin/destination ports + the list of visited links.
	 * 
	 * @param link
	 *            the link that is analyzed
	 * @param visitedLinks
	 *            the list of links already analyzed.
	 *            <em>This list is updated by the method.</em>
	 * @param tipPorts
	 *            the list of origin/destination ports.
	 *            <em>This list is updated by the method.</em>
	 */
	private static void extractConnectedPhysicalLinks(PhysicalLink link, Set<PhysicalLink> visitedLinks, Set<PhysicalPort> tipPorts) {
		// link is about to be analyzed. Add it to the list of visited links
		visitedLinks.add(link);

		// The principle of this method is to explore recursively each link
		// connected to the origin or destination port of link
		List<PhysicalLink> linksToVisit = new ArrayList<>();

		// First, determine for each port connected to link if it is a tip.
		// A port is a tip if its container either contains all containers of
		// the
		// other ports or none of them.
		for (AbstractPhysicalLinkEnd le : link.getLinkEnds()) {
			PhysicalPort p = (PhysicalPort) le;
			linksToVisit.addAll(port2link.get(p));

			Set<PhysicalComponent> portsParents = new HashSet<>();
			for (PhysicalLink l : port2link.get(p)) {
				PhysicalPort oppositePort;

				// Consider only the opposite port of p on link l
				if (l.getLinkEnds().get(0).equals(le)) {
					oppositePort = (PhysicalPort) l.getLinkEnds().get(1);
				} else {
					oppositePort = (PhysicalPort) l.getLinkEnds().get(0);
				}

				portsParents.add((PhysicalComponent) oppositePort.eContainer());
			}

			PhysicalComponent pParent = (PhysicalComponent) p.eContainer();
			boolean allParentsAreInsidePContainer = true;
			boolean allParentsAreOutsidePContainer = true;
			for (PhysicalComponent c : portsParents) {
				if (!EcoreUtil.isAncestor(pParent, c))
					allParentsAreInsidePContainer = false;
				else
					allParentsAreOutsidePContainer = false;
			}

			if (allParentsAreInsidePContainer || allParentsAreOutsidePContainer)
				tipPorts.add(p);
		}

		// Remove already visited links and continue exploration with neighbor
		// links
		linksToVisit.removeAll(visitedLinks);

		for (PhysicalLink l : linksToVisit) {
			extractConnectedPhysicalLinks(l, visitedLinks, tipPorts);
		}
	};

	/**
	 * Returns the path from root {@link PhysicalComponent} using '::' separator
	 * as the fully qualified name of <code>object</code>.
	 * <p>
	 * eg: <code>getFullQualifiedName(subsubsystem_3)</code> returns
	 * <code>Root::system_1::subsystem_2::subsubsystem_3</code>
	 * 
	 * @param object
	 *            the targeted object
	 * @return the fully qualified name of <code>object</code>
	 */
	public static String getFullQualifiedName(EObject object) {
		EObject parent = object.eContainer();
		if (!(parent instanceof PhysicalComponent))
			return ((NamedElement) object).getName();
		else
			return getFullQualifiedName((PhysicalComponent) parent) + "::" + ((NamedElement) object).getName();
	}


	/**
	 * Returns the instance path of <code>target</code> from origin {@link PhysicalComponent} using '.' separator.
	 * <p>
	 * eg: <code>getVirtualBehaviorLinkRelativeInstancePath(subsubsystem_3, root_system)</code> returns
	 * <code>system_1_part.subsystem_2_part.subsubsystem_3</code>
	 * 
	 * Used for <code>Actual_Connection_Binding</code> when the bus instance is not in the same container.
	 * 
	 * @param target
	 * @param origin
	 * @return
	 */
	public static String getVirtualBehaviorLinkRelativeInstancePath(CapellaElement target, PhysicalComponent origin) {
		EObject parent = null;
		if(target instanceof PhysicalLink) {
			Collection<PhysicalComponent> parents = ComponentExchangeUtils.getVirtualBehaviorLinkComponent((PhysicalLink) target);
			if(parents.size() == 1) {
				parent = (CapellaElement) parents.toArray() [0];
			}
		} else {
			parent = target.eContainer();
		}
		if (parent.equals(origin)) {
			return ((NamedElement) target).getName().toLowerCase().replace(' ', '_');
		} else {
			if (!(parent instanceof PhysicalComponent))
				return null; // TODO Improve error handling
			else
				return getVirtualBehaviorLinkRelativeInstancePath((PhysicalComponent) parent,
						origin) + "_part." + ((NamedElement) target).getName().toLowerCase().replace(' ', '_');
		}
	}

}
